const std = @import("std");
const ffi = @import("ffi.zig");
const c = ffi.c;

width: c_int,
height: c_int,
bit_depth: c_int,
/// This is cairo-friendly, not DBus
data: []u8,

const Self = @This();

/// Creates a new `Image` from data read from the dbus.
/// This will convert to a cairo-friendly format if needed.
/// `data` must be malloc'd
pub fn init(
    width: c_int,
    height: c_int,
    bit_depth: c_int,
    data: []u8,
) Self {
    const self = Self{
        .width = width,
        .height = height,
        .bit_depth = bit_depth,
        .data = data,
    };

    rgbaToCairo(self.data);

    return self;
}

pub fn createCairoSurface(self: *Self) !?*c.cairo_surface_t {
    // empty image
    if (self.data.len == 0)
        return null;

    const surface = c.cairo_image_surface_create_for_data(
        self.data.ptr,
        c.CAIRO_FORMAT_ARGB32,
        self.width,
        self.height,
        self.width * 4,
    ).?;
    errdefer c.cairo_surface_destroy(surface);
    try ffi.checkCairoStatus(c.cairo_surface_status(surface));
    return surface;
}

fn rgbaToCairo(data: []u8) void {
    std.debug.assert(data.len % 4 == 0);
    var i: usize = 0;
    while (i < data.len) : (i += 4) {
        const rgba = data[i .. i + 4];

        switch (@import("builtin").cpu.arch.endian()) {
            .Little => {
                std.mem.copy(
                    u8,
                    rgba,
                    &premultiply(.{
                        rgba[2],
                        rgba[1],
                        rgba[0],
                    }, rgba[3]),
                );
            },
            // TODO: somehow test this on big endian
            // (there's literally no way i got this right)
            .Big => {
                std.mem.copy(
                    u8,
                    rgba[1..],
                    &premultiply(.{
                        rgba[0],
                        rgba[1],
                        rgba[2],
                    }, rgba[3]),
                );
            },
        }
    }
}

// Premultiplies the rgb values with alpha
fn premultiply(rgb: [3]u8, alpha: u8) [3]u8 {
    const rgb16: @Vector(3, u16) = .{
        @intCast(rgb[0]),
        @intCast(rgb[1]),
        @intCast(rgb[2]),
    };

    // math!
    const z = rgb16 * @as(@Vector(3, u16), @splat(@intCast(alpha))) + @as(@Vector(3, u16), @splat(0x80));
    const z1 = z + (z >> @as(@Vector(3, u4), @splat(8))) >> @as(@Vector(3, u4), @splat(8));
    return .{
        @intCast(z1[0]),
        @intCast(z1[1]),
        @intCast(z1[2]),
    };
}
