const std = @import("std");
const ffi = @import("ffi.zig");
const c = ffi.c;
const State = @import("State.zig");
const Room = @import("Room.zig");
const Image = @import("Image.zig");
const IconStore = @import("IconStore.zig");

pub fn onRoomsReceived(proxy: *c.nhdbus, res: *c.GAsyncResult, s: *State) void {
    var rooms: ?*c.GVariant = null;
    var err: ?*c.GError = null;
    _ = c.nhdbus__call_rooms_finish(proxy, &rooms, res, &err);
    ffi.checkGError(err) catch return;
    std.log.info("Received rooms", .{});

    (std.Thread.spawn(.{}, readRooms, .{ rooms.?, s }) catch
        @panic("Failed to spawn room reader thread")).detach();
}

fn readRooms(rooms: *c.GVariant, s: *State) void {
    s.rooms_lock.lock();
    defer s.rooms_lock.unlock();

    const arena = s.room_arena.allocator();

    defer c.g_variant_unref(rooms);

    var iter = c.g_variant_iter_new(rooms);
    defer c.g_variant_iter_free(iter);

    s.rooms.ensureUnusedCapacity(
        @intCast(c.g_variant_iter_n_children(iter)),
    ) catch return;

    var room: ?*c.GVariant = null;
    while (true) {
        room = c.g_variant_iter_next_value(iter);
        if (room == null)
            break;
        defer c.g_variant_unref(room);

        var room_id: [*c]c.gchar = undefined;
        var room_alias: [*c]c.gchar = undefined;
        var room_name: [*c]c.gchar = undefined;
        var icon_url: [*c]c.gchar = undefined;
        var room_notifications: c_int = undefined;

        c.g_variant_get(
            room,
            "(&s&s&s&si)",
            &room_id,
            &room_alias,
            &room_name,
            &icon_url,
            &room_notifications,
        );

        const icon_url_s = std.mem.span(icon_url);

        const r = Room{
            .id = arena.dupeZ(u8, std.mem.span(room_id)) catch return,
            .alias = arena.dupeZ(u8, std.mem.span(room_alias)) catch return,
            .name = arena.dupeZ(u8, std.mem.span(room_name)) catch return,
            .icon_url = if (icon_url_s.len == 0)
                null
            else
                arena.dupeZ(u8, icon_url_s) catch return,
            .notifications = room_notifications,
        };

        s.rooms.append(r) catch return;
    }

    std.mem.sortUnstable(Room, s.rooms.items, {}, compareRoom);
    ffi.rofi_view_reload();
    std.log.info("Read {} rooms", .{s.rooms.items.len});
}

const IconUdata = struct {
    url: [:0]u8,
    store: *IconStore,
    data_alloc: std.mem.Allocator,
};

pub fn requestIcon(
    proxy: *c.nhdbus,
    store: *IconStore,
    url: [:0]const u8,
    data_alloc: std.mem.Allocator,
) !void {
    const url_d = try std.heap.c_allocator.dupeZ(u8, url);
    var udata = try std.heap.c_allocator.create(IconUdata);
    udata.* = .{
        .url = url_d,
        .store = store,
        .data_alloc = data_alloc,
    };

    c.nhdbus__call_image(
        proxy,
        url.ptr,
        null,
        @ptrCast(&onIconReceived),
        udata,
    );
}

fn onIconReceived(proxy: *c.nhdbus, res: *c.GAsyncResult, udata: *IconUdata) void {
    defer std.heap.c_allocator.destroy(udata);
    defer std.heap.c_allocator.free(udata.url);

    var image: ?*c.GVariant = null;
    var err: ?*c.GError = null;
    _ = c.nhdbus__call_image_finish(proxy, &image, res, &err);
    ffi.checkGError(err) catch return;

    defer c.g_variant_unref(image);

    var img_width: c_int = undefined;
    var img_height: c_int = undefined;
    var img_stride: c_int = undefined;
    var img_has_alpha: c_int = undefined;
    var img_bit_depth: c_int = undefined;
    var data_iter: *c.GVariantIter = undefined;

    c.g_variant_get(
        image,
        "(iiibiiay)",
        &img_width,
        &img_height,
        &img_stride,
        &img_has_alpha,
        &img_bit_depth,
        @as(?*anyopaque, null), // channel count
        &data_iter,
    );

    const data = udata.data_alloc.alloc(
        u8,
        if (img_has_alpha != 0)
            @intCast(c.g_variant_iter_n_children(data_iter))
        else
            (@as(usize, @intCast(c.g_variant_iter_n_children(data_iter))) / 3) * 4,
    ) catch return;

    var byte: u8 = 0;
    var i: usize = 0;

    if (img_has_alpha != 0) {
        while (c.g_variant_iter_next(data_iter, "y", &byte) != 0) {
            data[i] = byte;
            i += 1;
        }
    } else {
        while (c.g_variant_iter_next(data_iter, "y", &byte) != 0) {
            data[i] = byte;
            i += 1;
            if (i % 4 == 3) {
                data[i] = 0xff;
                i += 1;
            }
        }
    }

    const img = Image.init(
        img_width,
        img_height,
        img_bit_depth,
        data,
    );

    (udata.store.data.getOrPut(udata.url) catch return).value_ptr.* = .{ .cached = img };
    ffi.rofi_view_reload();
}

fn compareRoom(_: void, lhs: Room, rhs: Room) bool {
    if (lhs.notifications == rhs.notifications) {
        return std.ascii.lessThanIgnoreCase(
            std.mem.sliceTo(lhs.name, 0),
            std.mem.sliceTo(rhs.name, 0),
        );
    } else {
        return lhs.notifications > rhs.notifications;
    }
}
