const std = @import("std");
const ffi = @import("ffi.zig");
const c = ffi.c;
const dbus = @import("dbus.zig");
const Mode = c.Mode;
const Image = @import("Image.zig");
const State = @import("State.zig");
const Room = @import("Room.zig");
const IconStore = @import("IconStore.zig");

pub const std_options = struct {
    pub const logFn = @import("glib-log").log("rofi-nheko", 1024 * 4);
    // log levels are handled by glib, so zig shouldn't filter
    pub const log_level = .debug;
};

export var mode: Mode = .{
    .abi_version = c.ABI_VERSION,

    // elegant much
    .name = &struct {
        var name = "nheko".*;
    }.name,
    .cfg_name_key = [_]u8{0} ** 128,
    .display_name = null,
    .type = c.MODE_TYPE_SWITCHER,

    ._init = init,
    ._destroy = deinit,
    ._get_num_entries = getNumEntries,
    ._result = result,
    ._token_match = tokenMatch,
    ._get_display_value = getDisplayValue,
    ._get_icon = getIcon,
    ._create = null,
    ._completer_result = null,
    ._get_completion = null,
    ._preprocess_input = null,
    ._get_message = null,

    .free = null,
    .ed = null,
    .module = null,
    .fallback_icon_not_found = 0,
    .fallback_icon_fetch_uid = 0,

    .private_data = null,
};

fn init(m: [*c]Mode) callconv(.C) c_int {
    // For some incredibly weird reason,
    // rofi sometimes calls init and destroy multiple times.
    // This obviously causes all sorts of damage.
    if (c.mode_get_private_data(m) != null) {
        std.log.err("init called twice!", .{});
        return 1;
    }

    const state_p = std.heap.c_allocator.create(State) catch return 0;

    var err: ?*c.GError = null;

    const proxy = c.nhdbus__proxy_new_for_bus_sync(
        c.G_BUS_TYPE_SESSION,
        c.G_DBUS_PROXY_FLAGS_NONE,
        "im.nheko.Nheko",
        "/",
        null,
        &err,
    );
    ffi.checkGError(err) catch return 0;
    state_p.dbus = proxy.?;

    state_p.* = .{
        .dbus = proxy.?,
        .rooms = std.ArrayList(Room).init(std.heap.c_allocator),
        .rooms_lock = .{},
        .room_arena = std.heap.ArenaAllocator.init(std.heap.c_allocator),
        .icon_store = IconStore.init(std.heap.c_allocator),
    };

    c.mode_set_private_data(m, state_p);

    c.nhdbus__call_rooms(
        proxy,
        null,
        @ptrCast(&dbus.onRoomsReceived),
        state(m),
    );
    std.log.info("Requested rooms", .{});

    return 1;
}

fn deinit(m: [*c]Mode) callconv(.C) void {
    if (c.mode_get_private_data(m) == null) {
        std.log.err("destroy called twice!", .{});
        return;
    }
    defer c.mode_set_private_data(m, null);

    state(m).deinit();
}

fn getNumEntries(m: [*c]const Mode) callconv(.C) c_uint {
    state(m).rooms_lock.lock();
    defer state(m).rooms_lock.unlock();

    return @intCast(state(m).rooms.items.len);
}

fn result(
    m: [*c]Mode,
    mretv: c_int,
    _: [*c][*c]u8, // input
    selected_line: c_uint,
) callconv(.C) c.ModeMode {
    const l = std.log.scoped(.result);
    state(m).rooms_lock.lock();
    defer state(m).rooms_lock.unlock();

    l.debug("mretv = {x}", .{mretv});

    // no line selected
    if (selected_line == std.math.maxInt(c_uint) or (mretv & c.MENU_OK) == 0) {
        std.log.warn("Exiting without changing room", .{});
        return c.MODE_EXIT;
    }

    const room = state(m).rooms.items[selected_line];
    var err: ?*c.GError = null;
    _ = c.nhdbus__call_activate_room_sync(
        state(m).dbus,
        room.id.ptr,
        null,
        &err,
    );
    ffi.checkGError(err) catch {};
    l.info("Changed room to {s} ({s})", .{ room.id, room.name });

    // on custom action, switch room but don't exit
    if ((mretv & c.MENU_CUSTOM_ACTION) != 0) {
        l.info("Switched room without exiting", .{});
        return c.RELOAD_DIALOG;
    }
    return c.MODE_EXIT;
}

fn tokenMatch(
    m: [*c]const Mode,
    tokens: [*c][*c]c.rofi_int_matcher,
    index: c_uint,
) callconv(.C) c_int {
    state(m).rooms_lock.lock();
    defer state(m).rooms_lock.unlock();

    return c.helper_token_match(tokens, state(m).rooms.items[index].name.ptr);
}

fn getDisplayValue(
    m: [*c]const Mode,
    selected_line: c_uint,
    display_state: [*c]c_int,
    _: [*c][*c]c.GList, // attribute_list
    get_entry: c_int,
) callconv(.C) [*c]u8 {
    if (get_entry == 0)
        return null;

    state(m).rooms_lock.lock();
    defer state(m).rooms_lock.unlock();

    const room = state(m).rooms.items[selected_line];

    if (room.notifications > 0) {
        display_state.* |= 8; // STATE_MARKUP
        return (std.fmt.allocPrintZ(
            std.heap.c_allocator,
            "<span color='red'>({})</span> {s}",
            .{ room.notifications, room.name },
        ) catch @panic("Failed to allocate display value")).ptr;
    } else return c.g_strdup(room.name.ptr);
}

fn getIcon(
    m: [*c]const Mode,
    selected_line: c_uint,
    height: c_uint,
    scale: c_uint,
) callconv(.C) ?*c.cairo_surface_t {
    _ = scale;
    state(m).rooms_lock.lock();
    defer state(m).rooms_lock.unlock();

    var img = (state(m).icon_store.get(
        state(m).rooms.items[selected_line].icon_url orelse return null,
        state(m).dbus,
        state(m).room_arena.allocator(),
    ) catch return null) orelse return null;

    const img_surface = (img.createCairoSurface() catch return null) orelse return null;
    defer c.cairo_surface_destroy(img_surface);
    const resized_surface = c.cairo_image_surface_create(
        c.CAIRO_FORMAT_ARGB32,
        @intCast(height),
        @intCast(height),
    );
    ffi.checkCairoStatus(c.cairo_surface_status(resized_surface)) catch {
        c.cairo_surface_destroy(resized_surface);
        return null;
    };

    const cr = c.cairo_create(resized_surface);
    ffi.checkCairoStatus(c.cairo_status(cr)) catch return null;

    c.cairo_scale(
        cr,
        @as(f64, @floatFromInt(height)) / @as(f64, @floatFromInt(img.width)),
        @as(f64, @floatFromInt(height)) / @as(f64, @floatFromInt(img.height)),
    );
    c.cairo_set_source_surface(cr, img_surface, 0.0, 0.0);
    c.cairo_paint(cr);
    return resized_surface;
}

inline fn state(m: [*c]const Mode) *State {
    const p = c.mode_get_private_data(m);
    return @ptrCast(@alignCast(p));
}
